#include "planner.h"
#include "node.h"
#include <iostream>
#include <vector>
using namespace std;
using namespace __gnu_cxx;

Planner::Planner(std::vector<Action> actions, Literals init, Literals goal){
	_init = init;
	_goal = goal;
	_start = new StartNode(_init);
	_finish = new EndNode(_goal);
	addNode(_start);

	for(vector<Action>::iterator action = actions.begin(); action != actions.end(); ++action){
		Action* act = new Action(*action);
		_actions.push_back(act);
		const Literals& effects = act->effects(0);
		for (Literals::const_iterator effect = effects.begin(); effect != effects.end(); ++effect){
			cerr << "Adding effect: '" << *effect << "', action: " << action->name() << endl;
			_actionEffects[*effect] = act;
		}
	}
	cerr << "Number of actions: " << _actions.size() << endl;
	makePlan(_finish);
}

Planner::~Planner(){
	cerr << "Deleting " << _addedNodes.size() << " nodes" << endl;
	for (vector<Node*>::iterator node = _addedNodes.begin(); node != _addedNodes.end(); ++node){
		cerr << "Deleting node " << (*node)->action()->name() << endl;
		delete *node;
	}
	for (vector<Action*>::iterator action = _actions.begin(); action != _actions.end(); ++action){
		cerr << "Deleting action " << (*action)->name() << endl;
		delete *action;
	}
}


void Planner::makePlan(Node* node){
	addNode(node);

	cerr << "Fetching preconditions for action: " << node->action()->name() << ".. ";
	const Preconditions& preconds = node->action()->preconditions();
	cerr << "done" << endl;


	if (preconds.size() == 0){
		cerr << "Found no preconds" << endl;
		_start->addChild(node);
	}else{
		for (Preconditions::const_iterator precond = preconds.begin(); precond != preconds.end(); ++precond){
			cerr << "Looking for: '" << precond->first << "'" <<  endl;
			hash_map<string,Node*>::iterator addedNode = _addedEffects.find(precond->first);
			if(addedNode != _addedEffects.end()){
				cerr << "Using already added node" << endl;
				addedNode->second->addChild(node);
			}else {
				hash_map<string, Action*>::iterator action = _actionEffects.find(precond->first);
				if (action != _actionEffects.end()){
					cerr << "Adding new node" << endl;
					Node* newnode = new Node(action->second);
					newnode->addChild(node);
					makePlan(newnode);
				}else if (precond->second){
					cerr << "Action with effect: " << precond->first << " not found!" << endl;
					cerr << "This is a hard precondition, so this action and the children can't be executed." << endl;
					return;
				}else{
					cerr << "Action with effect: " << precond->first << " not found!" << endl;
					cerr << "This is a soft precondition, so we will continue" << endl;
					node->satisfyCondition(precond->first);
					_start->addChild(node);
				}
			}
		}
	}
}

void Planner::addNode(Node* node){
	cerr << "Adding node for action: " << node->action()->name() << endl;
	const Literals& effects = node->action()->effects(0);
	cerr << "Number of effects: " << effects.size() << endl;
	_addedNodes.push_back(node);

	for (Literals::const_iterator effect = effects.begin(); effect != effects.end(); ++effect){
		cerr << "Adding node for effect: " << *effect << endl;
		_addedEffects[*effect] = node;
	}
}


void Planner::execute(){
	_start->execute(Literals());
	cerr << "Number of nodes: " << _addedNodes.size() << endl;
	back_insert_iterator<Literals> ii(_init);
	_init.clear();
	int executions = 0;
	for (vector<Node*>::iterator node = _addedNodes.begin(); node != _addedNodes.end(); ++node){
		cerr << "Deleting node " << (*node)->action()->name() << endl;
		if ((*node)->executed()){
			executions++;
			const Literals& effects = (*node)->effects();
			copy(effects.begin(),effects.end(),ii);
			cerr << "Finding action" << endl;
			vector<Action*>::iterator action = find(_actions.begin(), _actions.end(), (*node)->action());
			if (action != _actions.end()){
				cerr << "Removing executed action: " << (*action)->name() << endl;
				delete *action;
				_actions.erase(action);
			}
		}
		delete *node;
	}
	_addedNodes.clear();
	_addedEffects.clear();
	_actionEffects.clear();
	copy(_init.begin(), _init.end(), ostream_iterator<string>(cerr, " "));
	cerr << endl;
	cerr << "Number of actions left: " << _actions.size() << endl;
	//TODO: Fill _actionEffects with the remaining effects, create start end end nodes and create a new plan.
	if (executions <= 1){
		cerr << "Non of the remaining actions could be executed, quiting." << endl;
		return;
	}
	if (_actions.size() == 0){
		cerr << "No remaining actions, quiting." << endl;
		return;
	}
	for (vector<Action*>::iterator action = _actions.begin(); action != _actions.end(); ++action){
		const Literals& effects = (*action)->effects(0);
		for (Literals::const_iterator effect = effects.begin(); effect != effects.end(); ++effect){
			cerr << "Adding effect: '" << *effect << "', action: " << (*action)->name() << endl;
			_actionEffects[*effect] = *action;
		}
	}
	_start = new StartNode(_init);
	_finish = new EndNode(_goal);
	addNode(_start);
	makePlan(_finish);
	if (_addedNodes.size() <= 2){
		cerr << "No actions to execute, quiting." << endl;
		return;
	}
	execute();
	
}
